        DEVICE ZXSPECTRUM48         ; Device setting for sjasmplus (.tap writing etc)
SCREEN  EQU $4000                   ; Location of screen
COLOR   EQU $5800                   ; Location of color array

        ORG $8000                   ; Let's start our code at 32k
        di                          ; Disable interrupts
        ld  sp,     0x8000          ; Set stack to grow down from our code
        ld  de,     0xfe00          ; im2 vector table start right after color table
        ld  hl,     0xfdfd          ; where interrupt will point at
        ld  a,      d 
        ld  i,      a               ; interrupt will hop to 0xfe?? where ?? is random 
        ld  a,      l               ; we need 257 copies of the address
rep_isr_setup:
        ld  (de),   a 
        inc e
        jr  nz,     rep_isr_setup
        inc d                       ; just one more
        ld  (de),   a
        ld de,      isr
        ld  (hl),   0xc3            ; 0xc3 = JP
        inc hl
        ld  (hl),   e
        inc hl
        ld  (hl),   d
        im  2                       ; set the interrupt mode
        ei                          ; Enable interrupt

        ; Entry: key select

        ld a, 0
        out (0xfe), a

        ld de, SCREEN
        ld hl, keyselect_scr
        ld bc, (256/8)*192+32*24
        ldir

inputselectloop:        
        call scaninputs        
        ld a, (keydata + 3)
        bit 0, a
        jr z, select_1
        bit 1, a
        jr z, select_2
        bit 2, a
        jr z, select_3
        jp inputselectloop
select_1:
        ld a, 0
        ld (inputmode), a
        jp levelselect        
select_2:
        ld a, 1
        ld (inputmode), a
        jp levelselect        
select_3:
        ld a, 2
        ld (inputmode), a
        jp levelselect        
        

levelselect:
        ld a, 0
        ld (movekey), a

        ld a, 0
        out (0xfe), a
        
        ; clear screen
        ld hl, SCREEN
        ld (hl), 0
        ld de, SCREEN + 1
        ld bc, (256/8)*192+32*24 - 1
        ldir

        ld de, map
        ld hl, levels
        ld bc, 16*12
        ldir
        
        call drawminimap
levelselectloop:
        call scaninputs
        
        ld a, (movekey)
        bit 4, a
        jr z, levelselectloop


        ld a, 0
        ld (movekey), a

        call findplayer
        ld (playerpos), hl
        call dirtymap
mainloop:
        ld a, 2
        out (0xfe), a

        call scaninputs
        call moveplayer

        ld hl, map + (16 * 12 - 1)
        ld bc, 0x100c ; 16x12
maploop:
        ld a, (hl)
        rla
        jr nc, skipdraw
        push hl
        push bc

        srl a
        ld (hl), a
        ld l, a
        ld h, 0
        
        push hl        
        ld hl, bc
        ld de, -0x0101
        add hl, de
        ld bc, hl
        pop hl
        
        ld a, 1
        out (0xfe), a
        call drawtile
        ld a, 2
        out (0xfe), a
        pop bc
        pop hl

skipdraw:        
        dec hl
        dec b
        jp nz, maploop
        ld b, 0x10
        dec c
        jr nz, maploop
        
        ld a, 4
        out (0xfe), a

        call physics

        ld a, 0
        out (0xfe), a
        
        halt

        jp mainloop

physofs:
        db 0, 160

; physics, calculates physics of non-player controlled items
; no parameters, destroys a, bc, ix (eventually others)        
physics:
        ld ix, map
        ld bc, (physofs)
        ld hl, -16
        add hl, bc
        ld a, l
        cp -16
        jr nz, physics_notlooped
        ld hl, 160
physics_notlooped:        
        ld (physofs), hl
        add ix, bc
        ld b, 16
physicsloop:
        ld a, (ix)
        cp 5
        jr z, physics_drop
        cp 6
        jr z, physics_drop
physicsdone:
        inc ix
        dec b
        ret z
        jp physicsloop
physics_drop:
        ld a, (ix+16)
        cp 5
        jr z, dropdiagonal_left
        cp 6
        jr z, dropdiagonal_left
        cp 0
        jr nz, physicsdone
        ld a, (ix)
        or 0x80
        ld (ix), 0x80
        ld (ix+16), a
        jr physicsdone
dropdiagonal_left:
        ld a, ixl   ; Note: undocumented instruction
        and 15
        jr z, dropdiagonal_right ; Can't move, border
        ld a, (ix-1)
        cp 0
        jr nz, dropdiagonal_right ; can't move, stuff on the left
        ld a, (ix+15)
        cp 0
        jr nz, dropdiagonal_right ; can't move, stuff at target
        ld a, (ix)
        or 0x80
        ld (ix), 0x80
        ld (ix+15), a
        jr physicsdone
dropdiagonal_right:
        ld a, ixl
        and 15
        cp 15
        jr z, physicsdone ; can't move, border
        ld a, (ix+1)
        cp 0
        jr nz, physicsdone ; can't move, stuff on the right
        ld a, (ix+17)
        cp 0
        jr nz, physicsdone ; cen't move, stuff at target
        ld a, (ix)
        or 0x80
        ld (ix), 0x80
        ld (ix+17), a
        jr physicsdone
         

; Draw mini-map for the level selection
drawminimap:
        ld hl, map + (16 * 12 - 1)
        ld bc, 0x100c ; 16x12
minimaploop:
        ld a, (hl)
        push hl
        push bc

        ld l, a
        ld h, 0
        
        push hl        
        ld hl, bc
        ld de, 0x0801
        add hl, de
        ld bc, hl
        pop hl
                
        call drawminitile
        pop bc
        pop hl

        dec hl
        dec b
        jp nz, minimaploop
        ld b, 0x10
        dec c
        jr nz, minimaploop
        ret

; drawminitile, copies tile data to screen
; input: hl = tile, b = x, c = y
; destroys de, hl, bc, a
drawminitile:
        ; Save these for later when we plot color
        push hl
        push bc
        ; One tile is 2*16 bytes
        add hl, hl ; *2
        add hl, hl ; *4
        add hl, hl ; *8
        add hl, hl ; *16
        add hl, hl ; *32
        ld de, tiles
        add hl, de
        ld de, hl  ; hl now is pointing at the start of tile x
        
        ; Figuring out the screen coordinates is trickier;
        ; screen coordinate bits go like this:
        ;           H          |            L
        ; 0 1 0 Y7 Y6 Y2 Y1 Y0 | Y5 Y4 Y3 X4 X3 X2 X1 X0
        ; rotate the coordinate right three bits (to get to Y3, Y4, Y5 in L)
        ld a, c
        rrca
        rrca
        rrca        
        ; AND any additional bits off
        and 0xe0
        ; Add x
        add a, b        
        ld l, a  ; coordinate bottom byte done
        ; next we do the same for Y6 and Y7; no need to shift because we're in the
        ; right place.
        ld a, c
        and 0x18 ; AND extra bits off, and h is done.
        ld h, a
        ld bc, SCREEN
        add hl, bc  ; now hl points at the screen offset we want
        
        ; Instead of looping, we'll plot each pixel separately..
        DUP 7
            ld a, (de)    ; Read pixels from data
            ld (hl), a    ; Write to screen
            inc de        ; Increment de and hl..
            inc de
            inc h
        EDUP
            ld a, (de)    ; Read pixels from data
            ld (hl), a    ; Write to screen
        
        ; Bitmap done, color to do
        pop bc
        ld l, c   ; y coordinate
        ld h, 0
        ; we need to multiply y by 32; 32 colors per scanline
        add hl, hl ; x2
        add hl, hl ; x4
        add hl, hl ; x8
        add hl, hl ; x16
        add hl, hl ; x32
        ld c, b
        ld b, 0
        add hl, bc ; x offset
        ld bc, COLOR
        add hl, bc
        ld bc, hl ; bc now has color table offset
        
        pop hl ; tile index
        ; One tile is 4 bytes of color
        add hl, hl ; x2
        add hl, hl ; x4
        ld de, tilecolors
        add hl, de  ; hl now points at tile color
        
        ld de, hl  ; de now points at tile
        ld hl, bc  ; hl now points at screen
        
        ld a, (de)      ; read color
        ld (hl), a      ; write color
        
        ret


; drawtile, copies tile data to screen
; input: hl = tile, b = x, c = y
; destroys de, hl, bc, a
drawtile:
        ; Save these for later when we plot color
        push hl
        push bc
        ; One tile is 2*16 bytes
        add hl, hl ; *2
        add hl, hl ; *4
        add hl, hl ; *8
        add hl, hl ; *16
        add hl, hl ; *32
        ld de, tiles
        add hl, de
        ld de, hl  ; hl now is pointing at the start of tile x
        
        ; Figuring out the screen coordinates is trickier;
        ; screen coordinate bits go like this:
        ;           H          |            L
        ; 0 1 0 Y7 Y6 Y2 Y1 Y0 | Y5 Y4 Y3 X4 X3 X2 X1 X0
        ; First, let's increment the input coordiate by x2 because
        ; we want 16 pixel step:
        ld a, c
        add a, a
        ld c, a 
        ; next, rotate the coordinate right three bits (to get to Y3, Y4, Y5 in L)
        rrca
        rrca
        rrca        
        ; AND any additional bits off
        and 0xe0
        ; Add in the x offset twice for 16 pixel step
        add a, b
        add a, b        
        ld l, a  ; coordinate bottom byte done
        ; next we do the same for Y6 and Y7; no need to shift because we're in the
        ; right place.
        ld a, c
        and 0x18 ; AND extra bits off, and h is done.
        ld h, a
        ld bc, SCREEN
        add hl, bc  ; now hl points at the screen offset we want
        
        ; Instead of looping, we'll plot each pixel separately..
        ld bc, 255
        DUP 7
            ld a, (de)    ; Read pixels from data
            ld (hl), a    ; Write to screen
            inc de        ; Increment de and hl..
            inc hl
            ld a, (de)    ; And repeat
            ld (hl), a
            inc de
            add hl, bc    ; Add in bc to move to the next line in screen (and one byte back)
        EDUP
            ld a, (de)    ; Read pixels from data
            ld (hl), a    ; Write to screen
            inc de        ; Increment de and hl..
            inc hl
            ld a, (de)    ; And repeat
            ld (hl), a
            inc de

        ld bc, 255 - 256 * 8 + 32 ; Move to the next block of 8 pixels
        add hl, bc
        ld bc, 255        ; And repeat the above process
        DUP 7
            ld a, (de)
            ld (hl), a
            inc de
            inc hl
            ld a, (de)
            ld (hl), a
            inc de
            add hl, bc            
        EDUP
            ld a, (de)
            ld (hl), a
            inc de
            inc hl
            ld a, (de)
            ld (hl), a
        
        ; Bitmap done, color to do
        pop bc
        ld l, c   ; y coordinate
        ld h, 0
        ; we need to multiply y by 64; 32 colors per scanline and we do 16x16 tiles
        add hl, hl ; x2
        add hl, hl ; x4
        add hl, hl ; x8
        add hl, hl ; x16
        add hl, hl ; x32
        add hl, hl ; x64
        ld c, b
        ld b, 0
        add hl, bc ; x offset
        add hl, bc ; x offset x2, for 16x16 tiles
        ld bc, COLOR
        add hl, bc
        ld bc, hl ; bc now has color table offset
        
        pop hl ; tile index
        ; One tile is 4 bytes of color
        add hl, hl ; x2
        add hl, hl ; x4
        ld de, tilecolors
        add hl, de  ; hl now points at tile color
        
        ld de, hl  ; de now points at tile
        ld hl, bc  ; hl now points at screen
        ld bc, 31  ; scanline - 1
        
        ld a, (de)      ; read color
        ld (hl), a      ; write color
        inc de          ; next color
        inc hl          ; next pixel
        ld a, (de)
        ld (hl), a
        inc de
        add hl, bc      ; next row - 1 pixel
        ld a, (de)
        ld (hl), a
        inc de
        inc hl
        ld a, (de)
        ld (hl), a       
        
        ret

; Marks all tiles as dirty and requiring redraw
; no inputs, destroys hl, a, b
dirtymap:
        ld hl, map
        ld b, (16*12)
dmloop:
        ld a, (hl)
        or 0x80
        ld (hl), a
        inc hl
        dec b
        jr nz, dmloop
        ret

; Marks one tile as dirty
; hl=tile index, destroys hl, bc, a
dirtytile:
        ld bc, map
        add hl, bc
        ld a, (hl)
        or 0x80
        ld (hl), a
        ret
        
; findplayer - scans map and returns the player's position
; no parameters, destroys a, bc, returns position in hl
findplayer:                
        ld hl, map - 1
        ld a, 1
        ld bc, 16*12
        cpir
        ld bc, -map-1
        add hl, bc
        ret

keydata:
        db 0,0,0,0,0,0,0,0,0,0
wasdown:
        db 0
movekey:
        db 0
inputmode:
        db 0
        
scaninputs:
        ld bc, 0xfefe
        in a, (c)
        ld (keydata + 0), a
        ld bc, 0xfdfe
        in a, (c)
        ld (keydata + 1), a
        ld bc, 0xfbfe
        in a, (c)
        ld (keydata + 2), a
        ld bc, 0xf7fe
        in a, (c)
        ld (keydata + 3), a
        ld bc, 0xeffe
        in a, (c)
        ld (keydata + 4), a
        ld bc, 0xdffe
        in a, (c)
        ld (keydata + 5), a
        ld bc, 0xbffe
        in a, (c)
        ld (keydata + 6), a
        ld bc, 0x7ffe
        in a, (c)
        ld (keydata + 7), a
        ld bc, 31 ; kempston
        in a, (c) 
        ld (keydata + 8), a

        MACRO SCANKEY keybyte, keybit, outbit
            ld a, (keydata + keybyte)
            bit keybit, a
            jr z, .scan_key_down
            ld a, (wasdown)
            and a, 0xff ^ (1 << outbit)
            ld (wasdown), a
            jr .scan_done
.scan_key_down:
            ld a, (wasdown)
            bit outbit, a
            jr nz, .scan_done
            or a, 1 << outbit
            ld (wasdown), a
            ld a, (movekey)
            or a, 1 << outbit
            ld (movekey), a
.scan_done:
        ENDM
        
        SCANKEY 7,0,4 ; space        
        
        ld a, 0
        ld hl, inputmode
        sub (hl)
        jr z, input_wasd
        ld a, 1
        sub (hl)
        jp z, input_qaop
        ld a, 2
        sub (hl)
        jp z, input_kempston
        ; fallthrough to wasd        
                
        ; 'W' : byte 2, bit 1
        ; 'A' : byte 1, bit 0
        ; 'S' : byte 1, bit 1
        ; 'D' : byte 1, bit 2
input_wasd:
        SCANKEY 2,1,0 ; up 
        SCANKEY 1,1,1 ; down
        SCANKEY 1,0,2 ; left
        SCANKEY 1,2,3 ; right        
        ret        
        
        ; 'Q' : byte 2, bit 0
        ; 'A' : byte 1, bit 0
        ; 'O' : byte 5, bit 1
        ; 'P' : byte 5, bit 0
input_qaop:
        SCANKEY 2,0,0 ; up 
        SCANKEY 1,0,1 ; down
        SCANKEY 5,1,2 ; left
        SCANKEY 5,0,3 ; right        
        ret        

        ; kempston up    : byte 8, bit 2
        ; kempston down  : byte 8, bit 3
        ; kempston left  : byte 8, bit 0
        ; kempston right : byte 8, bit 1
input_kempston:        
        SCANKEY 8,2,0 ; up 
        SCANKEY 8,3,1 ; down
        SCANKEY 8,0,2 ; left
        SCANKEY 8,1,3 ; right        
        ret        

moveplayer:
        ld a, (movekey)
        bit 4, a
        jp nz, levelselect


        ld hl, (playerpos)
        ld bc, map
        add hl, bc
        ld (hl), 0x80
        ld hl, (playerpos)

        ; Trying to move up, check if we're not on the top row
        ld a, l
        cp 16
        jr c, notup
        
        ld a, (movekey)
        bit 0, a
        jr z, notup
        ld bc, -16
        add hl, bc
        jp moved
notup:
        ; Moving down, check if we're not on the bottom row
        ld a, l
        cp 16*11
        jr nc, notdown
        
        ld a, (movekey)
        bit 1, a
        jr z, notdown
        ld bc, 16
        add hl, bc
        jp moved
notdown:
        ; Moving left, x coordinate may not be zero
        ld a, l
        and 0xf
        jr z, notleft
        
        ld a, (movekey)
        bit 2, a
        jr z, notleft
        dec hl
        jp moved
notleft:
        ; moving right, x coordinate may not be 15
        ld a, l
        and 0xf
        cp 15
        jr z, notright
        
        ld a, (movekey)
        bit 3, a
        jr z, notright
        inc hl
notright:
moved:
        ld a, 0
        ld (movekey), a
        
        ; Now that we have a *potential* new position in hl, we need to
        ; figure out if it's legal.
        ld bc, hl
        ld de, map
        add hl, de
        ld a, (hl)
        ld hl, bc
        
        cp 0
        jr z, moveok ; Empty tile is always fine
        cp 2
        jr z, moveok ; Ground is fine
        cp 5
        jr z, moveok ; Gems are fine
        cp 8
        jr z, moveok ; Open exit is fine
        cp 6
        jr z, movestone ; Check if we can push a stone
        jr movedone
        
movestone:
        ld a, (playerpos)
        sub l ; playerpos - target
        ; Let's limit stone pushing to horizontal movement
        ; -1 = 11111111, 1 = 00000001, 16 = 00010000, -16 = 11110000
        ; checking bit 0 is enough
        bit 0, a
        jr z, movedone
        ld bc, hl  ; back up target position
        ld l, a    ; l = 1/-1
        ld a, c    ; a = target position
        sub l      ; a -= 1/-1
        ld l, a    ; hl = other side of target position
        ld de, map
        add hl, de
        ld a, (hl)
        cp 0
        jr nz, movedone ; other side wasn't empty
canmovestone:
        ld (hl), 0x86 ; store stone in empty slot
        ld hl, bc  ; restore move target

        ; fallthrough        

moveok:
        ld (playerpos), hl
movedone:
        ld hl, (playerpos)
        ld bc, map
        add hl, bc
        ld (hl), 0x81        
        ret


isr:                    ; This will be called ~50Hz
        push bc
        ld   bc, (framecounter)
        inc  bc
        ld   (framecounter), bc
        pop bc
        ei
        reti            ; Return from interrupt

framecounter:
        db 0,0

playerpos:
        db 0,0


map:
        BLOCK 16*12,0
        
levels:        
        db 2,2,6,3,5,2,2,2,2,2,2,2,2,2,2,2
        db 2,5,6,3,2,2,6,2,2,2,2,2,2,2,2,2
        db 2,2,2,3,2,6,2,2,2,2,2,2,2,2,2,2
        db 3,2,3,3,2,2,2,6,2,2,2,2,2,2,2,2
        db 2,2,2,6,2,2,2,2,2,2,2,2,2,2,2,2
        db 2,2,0,2,2,2,6,2,2,2,2,2,2,2,2,2
        db 0,1,0,2,6,2,2,7,2,2,2,2,2,2,2,2
        db 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2
        db 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2
        db 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2
        db 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2
        db 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2

tiles:
        BLOCK 32,0 	; empty tile
        INCBIN "tiles.dat"

        ; 0x00 0x00 = black      0x04 0x20 = green
        ; 0x01 0x08 = blue       0x05 0x28 = cyan
        ; 0x02 0x10 = red        0x06 0x30 = yellow
        ; 0x03 0x18 = magenta    0x07 0x38 = white
        ;      0x40 = bright          0x80 = blink
tilecolors:
        db 0x00, 0x00, 0x00, 0x00 ; 0 empty
        db 0x05, 0x05, 0x04, 0x04 ; 1 player
        db 0x10, 0x10, 0x10, 0x10 ; 2 ground
        db 0x50, 0x50, 0x50, 0x50 ; 3 bricks
        db 0x04, 0x04, 0x04, 0x04 ; 4 goo
        db 0x45, 0x05, 0x05, 0x45 ; 5 gem
        db 0x07, 0x07, 0x07, 0x07 ; 6 stone
        db 0x04, 0x04, 0x04, 0x04 ; 7 exit (closed)
        db 0x44, 0x44, 0x44, 0x44 ; 8 exit (open)

keyselect_scr:
        INCBIN "keyselect.scr"            	
    	
    	SAVETAP "test.tap", $8000   ; Save the assembled program as a tap file
